#include "../../agtree.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __linux__
#include <getopt.h>
#endif
/* extern vars for getopt, parsing cmdline */
extern int optind;
extern char *optarg;

 

void parse_cmdline(struct cmds *cmds, int ac, char **av);
void cmd_err();


void main(int argc, char **argv)
{
  struct cmds params;
  
  parse_cmdline(&params,argc,argv);
  fflush(stdout);fflush(stderr);

  make_agtree(&params);
  fflush(stdout);fflush(stderr);
  
}

void cmd_err()
{
  fprintf(stderr,"\n\nUsage: make_agtree [-p prune_file] [-r regularization_factor (0.001)] [-l no_leaves (50)] [-m no_root nodes_after_merge (1)] train_file out_file\a\n\n");
  exit(-1);
}

void parse_cmdline(struct cmds *cmds, int ac, char **av)
{

  int i,ct;
  
  cmds->prune=0;
  cmds->alpha = 0.001;
  cmds->merge_limit = 1; 
  cmds->nb_clusters = 50;

  for (;;)
    {
      i= getopt(ac, av, "p:r:m:l:");
      if (i==EOF)
	break;
      else
	{
	  if (i == '?')
	    cmd_err();

	  /* pruning on, arg. is prune file */
	  if (i == 'p')
	    {
	      strcpy(cmds->prunefile, optarg);
	      cmds->prune=1;
	    }

	  if (i == 'l')
	    cmds->nb_clusters = atoi(optarg);

	  if (i == 'm')
	    cmds->merge_limit =  atoi(optarg);

	  /* tree-merging regularization factor, 
	     parameter alpha is argument */
	  if (i == 'r')
	    cmds->alpha = atof(optarg);
	    

	}
    }

  ct=optind;
  
  if (ac-optind != 2) 
    cmd_err();
  
  strcpy(cmds->trainfile, av[ct]);
  ct++;
  strcpy(cmds->outfile, av[ct]);
  ct++;

  /* end parse input */


  /* assign variables, check and print config */
  cmds->nb_nodes = cmds->nb_clusters; /* before merging, 
					 nb_nodes=nb_clusters */

  /* -m 0 switches off merging all together, output will be initial
     vector quantization, i.e. 'leaves' only */
  if ( cmds->merge_limit == 0)
    cmds->merge_limit = -1;
     
   
  /* regularization factor must be positive */
  if (cmds->alpha < 0.0) 
    {
      fprintf(stderr,"\n\nregularization factor must be greater than zero, reg-fac=%f\a\n\n",cmds->alpha);
      cmd_err();

      exit(-1);
    }

  /* check: if pruning is on, must merge to a single root node first*/
    /* Check input parameters */
  if ((cmds->prune==1) && (cmds->merge_limit != 1))
    {
      fprintf(stderr,"With pruning on, merge_limit must be 1, is %d\n\a\n",cmds->merge_limit);
      cmd_err();
      exit(-1);
    } 

  

  printf("infile: %s\noutfile: %s\nno_leaves: %d merge_limit=%d\n\n",cmds->trainfile,cmds->outfile,cmds->nb_clusters, cmds->merge_limit);


  if (cmds->prune == 1)
    printf("pruning on, prunefile %s\n",cmds->prunefile);
  else
    printf("pruning OFF.\n\n");
   




}


