#include "../../agtree.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


/* predict_vec is the important part - how to classify an input 
   vector x_vec to a tree node and use the prediction filter
   in that tree node to calculate y_vec */
void predict_vec(double *y_vec,double *x_vec,int x_len,int y_len,struct node *tree, struct cmds *params)
{
  int cl;
  int j,k;
  
  /* classify input vector x_vec to a specific tree node */
  /* classify_xvec is in libagt.a */
  cl=classify_xvec(x_vec, x_len, tree, params);

  /* This is how prediction covariance matrix is obtained,
     commented out for this example */
  /*for (i=0;i<y_len;i++)
      for (j=0;j<y_len;j++)
        y_covariance_matrix[i][j] = clust[cl].yer2[i][j];
  */
  
  
  /* prediction of y_vec from x_vec: y=Ax+b where A and b 
     are functions of tree node cl*/
  for (j=0;j<y_len;j++)
    y_vec[j] =0.0;
  for (j=0;j<y_len;j++)
    for (k=0;k<x_len;k++)
      y_vec[j] += (x_vec[k] * tree[cl].a[k][j]);
  
  
  /* add b, i.e. y_hat = A*x + b    */
  for (j=0;j<y_len;j++)
    y_vec[j] += tree[cl].b[j];
  
}




/* The main routine does mostly FILE I/O for this example
   and calls predict_vec for each data point */
int main(int argc,char **argv)
{
  struct node *tree;
  struct cmds params;

  char *tree_fname,*in_file,*out_file;
  int nb_clusters; /* number of leaves */
  int nb_nodes; /* number of tree nodes=2*no_clusters for full
		   binary tree */
  int xlen,ylen,xlen_input,no_pts_input,i,pt,stat;
  FILE *fp_input,*fp_output;
  double *x_vec,*y_vec;




  /* Get commandline parameters */
  if (argc!=4)
    {
      fprintf(stderr,"Usage: predict tree-file input-file output-file\n\a\n");
      exit(-1);
    }
  
  tree_fname=argv[1];
  in_file=argv[2];
  out_file=argv[3];




  /* read tree info */
  tree = read_tree(tree_fname, &xlen, &ylen, &nb_clusters, &nb_nodes);
  params.nb_clusters = nb_clusters;
  params.nb_nodes = nb_nodes;
  printf("\nTree parameters:\n");
  printf("  length of  input vector x-len: %d\n",xlen);
  printf("  length of output vector y-len: %d\n",ylen);
  printf("  number of leaves in tree: %d\n",nb_clusters);
  printf("  number of nodes incl. leaves: %d\n",nb_nodes);
  



  /* open input file and read header info*/
  fp_input = fopen(in_file,"r");
  if (fp_input==NULL)
    {
      fprintf(stderr,"Error: Can't open input file %s.\n\a\n",in_file);
      exit(-1);
    }
  fscanf(fp_input,"%d",&no_pts_input);
  fscanf(fp_input,"%d",&xlen_input);
  

  if (xlen_input!=xlen)
    {
      fprintf(stderr,"Error: input vector length of input file is incompatible with that of tree.\n\a\n");
      exit(-1);
    }

  

  /* allocate x and y vectors */
  x_vec=(double*)malloc(xlen*sizeof(double));
  y_vec=(double*)malloc(ylen*sizeof(double));
  if ((x_vec==NULL)||(y_vec==NULL))
    {
      fprintf(stderr,"Malloc Error.\n\a\n");
      exit(-1);
    }


  /* open output file */
  fp_output= fopen(out_file,"w");
  if (fp_output==NULL)
    {
      exit(-1);
      fprintf(stderr,"Error: Can't open/write output file %s.\n\a\n",out_file);
    }


  /* loop through data vectors */
  for (pt=0;pt<no_pts_input;pt++)
    {
      /* read vector from input file */
      for (i=0;i<xlen;i++)
	{
	  stat=fscanf(fp_input,"%lf",&x_vec[i]);
	  if ((stat!=1) || (stat==EOF))
	    {
	      fprintf(stderr,"Read error in input file.\n\a\n");
	      exit(-1);
	    }
	}
      
      /* call subroutine predict in this file for single data point */
      predict_vec(y_vec,x_vec,xlen,ylen,tree,&params);
      
      /* write predicted vector */
      for (i=0;i<ylen;i++)
	fprintf(fp_output,"%g ",y_vec[i]);
      fprintf(fp_output,"\n");

    }
  
  fclose(fp_input);
  fclose(fp_output);

  free((void*)x_vec);
  free((void*)y_vec);


  printf("Done.\n"); 
  return(0);
}




