#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "allocate.h"

#include "svdcmp.h"

static void matrixProd ( double **a1, int a1width,
                         double **a2, int a2height,
                         double **a3,
                         int a3height, int a3width );
static double computeTol ( double *w, int n );


void pseudoinvert ( double **a, int n )
{
  int i,j;
  double tol;
  static double **U,**Utrans,**V,*w,**Winv;
  static int prev_n,first=1;

  if(first){
    U=(double **)get_img(n+1,n+1,sizeof(double));
    Utrans=(double **)get_img(n+1,n+1,sizeof(double));
    V=(double **)get_img(n+1,n+1,sizeof(double));
    w=(double *)get_spc(n+1,sizeof(double));
    Winv=(double **)get_img(n+1,n+1,sizeof(double));

    prev_n = n;
    first=0;
  }
  if (n != prev_n){
    free_img((void **)(U));
    free_img((void **)(Utrans));
    free_img((void **)(V));
    free((void *)(w));
    free_img((void **)(Winv));

    U=(double **)get_img(n+1,n+1,sizeof(double));
    Utrans=(double **)get_img(n+1,n+1,sizeof(double));
    V=(double **)get_img(n+1,n+1,sizeof(double));
    w=(double *)get_spc(n+1,sizeof(double));
    Winv=(double **)get_img(n+1,n+1,sizeof(double));

    prev_n = n;
  }
  for(i=0;i<=n;i++) {
    for(j=0;j<=n;j++) U[i][j] = 0.0;
    for(j=0;j<=n;j++) V[i][j] = 0.0;
    w[i] = 0.0;
  }

  /* copy input matrix into U */
  for(i=1;i<=n;i++)
  for(j=1;j<=n;j++) U[i][j] = a[i-1][j-1];

  /* singular value decomposition */
  svdcmp(U,n,n,w,V);
  
  /*printf("\n\n\n");
  for (i=0;i<=n;i++)
    printf("%g  ",w[i]);
  printf("\n\n");
  */

  /* compute tolerance; similar to what matlab does */
  tol = computeTol(w,n);
  /*printf("tol=%g\n",tol);*/
  /* compute pseudoinverse as V*inv(W)*U^t, with elem's in */
  /* Winv corresponding to (zero elements in W) zeroed out */

  for(i=1;i<=n;i++)
  for(j=1;j<=n;j++){
    if(i==j){
      if   (w[i]<=tol) Winv[i][j] = 0.0;
      else             Winv[i][j] = 1.0/(w[i]);
    }
    else Winv[i][j] = 0.0;
  }

  for(i=1;i<=n;i++)
  for(j=1;j<=n;j++) Utrans[i][j] = U[j][i];

  matrixProd ( Winv, n, Utrans, n, U,      n, n );
  matrixProd ( V,    n, U,      n, Utrans, n, n );

  /* copy output matrix into a */
  for(i=1;i<=n;i++)
  for(j=1;j<=n;j++) a[i-1][j-1] = Utrans[i][j];
}

static double computeTol ( double *w, int n )
{
  int i;
  double max,tol;

  max = -1.0;

  for(i=1;i<=n;i++) if(w[i]>max) max=w[i];

  if(max<0.0){
    printf("problem computing svd - exiting\n");
    exit(1);
  }

  tol = ((double)(n)) * max * (1.00e-15);
  return(tol);
}

static void matrixProd ( double **a1, int a1width,
                         double **a2, int a2height,
                         double **a3,
                         int a3height, int a3width )
{
  /* a3 <-- a1 * a2 */

  int i,j,m;

  if ( a1width != a2height ) {
    printf("error:matrixProd / pseudoinvert:  matrix dim's do not agree\n");
    exit(1);
  }

  for(i=1;i<=a3height;i++)
  for(j=1;j<=a3width;j++) {
    a3[i][j] = 0.0;
    for(m=1;m<=a1width;m++) a3[i][j] += a1[i][m] * a2[m][j];
  }
}

