#include "../include/agt_merge.h"

/* This file contains code to merge initial vq clusters
   into a binary tree such that the initial clusters are the
   leaves of the tree */


static double abval_sq(double *vec,int x_len);    

void Merging(struct node *clust, struct data *train, struct cmds *params)
{
  struct heap heap;
  struct heap_node heap_node;
  int nb_clusters, cl1,cl2,i;
  int nb_root_nodes, merge_limit,ct;
  double mcost;
  
  nb_clusters = params->nb_clusters;
  merge_limit = params->merge_limit;

  /* Initially, nodes in the tree are the leaf clusters */
  params->nb_nodes = nb_clusters;
  /* and all nodes(=clusters) are root nodes */
  nb_root_nodes = nb_clusters;

  initMerging(&heap, clust, train, params);

  while ((heap.size > 0) && (nb_root_nodes > merge_limit))
    {
      get_heap_max(&heap, &heap_node);
      cl1=heap_node.pt/(2*nb_clusters);
      cl2=heap_node.pt-cl1*2*nb_clusters;
      /*printf("\nConsidering  cl1=%d cl2=%d mcost=%g rt1=%d rt2=%d\n",cl1,cl2,-1.0*heap_node.x,clust[cl1].root_FLAG,clust[cl2].root_FLAG);*/

      if ((clust[cl1].root_FLAG==1) && (clust[cl2].root_FLAG==1))
	{
	  mcost=-1.0*heap_node.x;
	  
	  printf("Merging cl1=%d cl2=%d mcost=%g\n",cl1,cl2,mcost);

	  /* merge cl1 and cl2 into clust[params->nb_nodes] */
	  merge(&clust[params->nb_nodes], clust, cl1, cl2, train); 
	  
	  /* now cl1, cl2 are no longer root nodes */
	  clust[params->nb_nodes].root_FLAG=1;
	  clust[params->nb_nodes].child1=cl1;
	  clust[params->nb_nodes].child2=cl2;


	  clust[cl1].root_FLAG=0;
	  clust[cl2].root_FLAG=0;
	  /* set parent indices (''pointers') in cl1 and cl2 */
	  clust[cl1].parent = params->nb_nodes;
	  clust[cl2].parent = params->nb_nodes;

	  /* Now add all new combinations of this node and the */
	  /* remaining root nodes to the heap */
	  for (i=0;i<params->nb_nodes;i++)
	    if (clust[i].root_FLAG==1)
	      {
		heap_node.pt = i*2*nb_clusters + params->nb_nodes;
		mcost = merge_cost(clust,i,params->nb_nodes, train,params);
		heap_node.x = -1.0*mcost;
		heap_insert(&heap, &heap_node);
		
	      }
	  
	  /* number of nodes increased by 1 */
	  params->nb_nodes = params->nb_nodes + 1;
	  printf("params->nb_nodes = %d\n",params->nb_nodes);
	  if (params->nb_nodes>2*nb_clusters)
	    {
	      fprintf(stderr,"\n\nError: params->nb_nodes>2*nb_clusters\a\n\n");
	      exit(-1);
	    }
	  /* and by merging two root nodes into one, we have reduced */
	  /* the number of root_nodes by 1 */
	  nb_root_nodes--;

	} 

    }
  printf("Done merging\n\n");
  /*printTree(clust, params);*/

  /* check nb root nodes */
  ct=0;
  for (i=0;i<params->nb_nodes;i++)
    if (clust[i].root_FLAG==1)
      ct++;

  /*printf("Counted root nodes=%d, calculated = %d\n",ct,nb_root_nodes);*/

  if (ct!=nb_root_nodes)
    {
      fprintf(stderr,"\n\nError: Counted root nodes=%d, calculated = %d\a\n\n",ct,nb_root_nodes);
      exit(-1);
    }

}







void initMerging(struct heap *heap, struct node *clust, struct data *train, struct cmds *params)
{
  int nb_clusters;
  int i,j;
  struct heap_node heap_node;
  double mcost;


  nb_clusters = params->nb_clusters;
 
  initialize_heap(heap);

 
  for (i=0;i<nb_clusters-1;i++)
    {
      for (j=i+1;j<nb_clusters;j++)
	{
	  heap_node.pt = i*2*nb_clusters + j;
	  mcost = merge_cost(clust,i,j, train,params);
	  heap_node.x = -1.0*mcost;
	  heap_insert(heap, &heap_node);
	}
    }
}





double merge_cost(struct node *clust, int cl1, int cl2, struct data *train, struct cmds *params)
{
  struct node *mgnode;
  double cost, cost_for_meandist=0.0; 
  static double ct=0.0;
  int x_len;

  double mu1_absq, mu2_absq, muN_absq;

  ct++;

  x_len=train->x_len;


  if ((clust[cl1].nb_points==0) || (clust[cl2].nb_points==0))
    {

      cost=0.0;

      if (clust[cl1].parent == clust[cl2].parent)
	{
	  printf("Setting mcost=-10.0 for clusters %d (%d points) and %d (%d points), one empty\n",cl1,clust[cl1].nb_points,cl2,clust[cl2].nb_points);
	  cost=-10.0;
	}
      else
	{
	  printf("Setting mcost=1.0 for clusters %d (%d points) and %d (%d points), one empty\n",cl1,clust[cl1].nb_points,cl2,clust[cl2].nb_points);
	  cost=1.0;
	}

    }
  else
    {

      mgnode = allocClusters(1, train->x_len, train->y_len);

      merge(mgnode, clust, cl1, cl2, train); 


      /* FIRST COST TERM is increase in prediction error */
      cost = mgnode->pred_err - (clust[cl1].pred_err + clust[cl2].pred_err);

      if (cost<1e-10)
	cost=0.0;

      
      if (params->alpha > 0.0)
	{
	  mu1_absq=abval_sq(clust[cl1].mean,x_len);
	  mu2_absq=abval_sq(clust[cl2].mean,x_len);
	  muN_absq=abval_sq(mgnode->mean,x_len);


	  /* SECOND, REGULARIZING COST TERM, ACCOUNTING FOR DISTANCE
	     OF THE CENTERS OF THE 2 CLUSTERS MERGED (distance calculated
	     in input space x)
	  */
	  cost_for_meandist = params->alpha * (clust[cl1].nb_points * mu1_absq + clust[cl2].nb_points * mu2_absq - mgnode->nb_points * muN_absq);

	  if (cost_for_meandist<0.0)
	    {
	      fprintf(stderr,"Error: cost_for_meandist=%g < 0.0\n\a\n",cost_for_meandist);
	      exit(-1);
	    }

	  if (mgnode->nb_points!=clust[cl1].nb_points+clust[cl2].nb_points)
	    {
	      fprintf(stderr,"Error: mgnode->nb_points!=clust[cl1].nb_points+clust[cl2].nb_points\n\a\n");
	      exit(-1);
	    }
	  
	  cost += cost_for_meandist;
	}
      
      params->avg_cost_md = (params->avg_cost_md*(ct-1.0) + cost_for_meandist)/ct;
      params->avg_cost_tot = (params->avg_cost_tot*(ct-1.0) + cost)/ct; 
      

      
      freeClusterStatsandFlt(mgnode);
      free((void*)mgnode->mean);
      free((void*)mgnode);
    } /* end else nb_points==0*/

  return(cost);


}




void merge(struct node *mgnode, struct node *clust, int cl1, int cl2, struct data *train)
{
  int x_len,y_len;
  int i,j;
  double weight1, weight2;

  x_len=train->x_len;
  y_len=train->y_len;

  /*allocClusterStats(mgnode, x_len, y_len);*/

  if ((clust[cl1].nb_points>0) || (clust[cl2].nb_points>0))
    {

      /* Now merge statistics */
      mgnode->nb_points=clust[cl1].nb_points + clust[cl2].nb_points;
      weight1 = ((double)clust[cl1].nb_points)/((double)mgnode->nb_points);
      weight2 = ((double)clust[cl2].nb_points)/((double)mgnode->nb_points);
      
      for (i=0;i<x_len;i++)
	{
	  mgnode->xmean[i] = weight1*clust[cl1].xmean[i] + weight2*clust[cl2].xmean[i];
	  mgnode->mean[i] = weight1*clust[cl1].mean[i] + weight2*clust[cl2].mean[i];
	}
      for (i=0;i<y_len;i++)
	mgnode->ymean[i] = weight1*clust[cl1].ymean[i] + weight2*clust[cl2].ymean[i];
      
      /* merge covariance Cxx */
      for (i=0;i<x_len;i++)
	for (j=0;j<x_len;j++)
	  mgnode->R[i][j] = weight1*(clust[cl1].R[i][j]+clust[cl1].xmean[i]*clust[cl1].xmean[j]) +  weight2*(clust[cl2].R[i][j]+clust[cl2].xmean[i]*clust[cl2].xmean[j]) - mgnode->xmean[i]*mgnode->xmean[j]; 
  
      /* merge cross-covariance Cxy */
      for (i=0;i<x_len;i++)
	for (j=0;j<y_len;j++)
	  mgnode->vect[i][j] = weight1*(clust[cl1].vect[i][j]+clust[cl1].xmean[i]*clust[cl1].ymean[j]) +  weight2*(clust[cl2].vect[i][j]+clust[cl2].xmean[i]*clust[cl2].ymean[j]) -  mgnode->xmean[i]*mgnode->ymean[j]; 
      
      
      /* merge covariance Cyy */
      for (i=0;i<y_len;i++)
	for (j=0;j<y_len;j++)
	  mgnode->Cyy[i][j] = weight1*(clust[cl1].Cyy[i][j]+clust[cl1].ymean[i]*clust[cl1].ymean[j]) +  weight2*(clust[cl2].Cyy[i][j]+clust[cl2].ymean[i]*clust[cl2].ymean[j]) - mgnode->ymean[i]*mgnode->ymean[j]; 
  
       
      /* Compute predictor and prediction error for new node */
      compute_clust_filter(mgnode, train);
      compute_clust_prederr(mgnode, train);
    }
  else /* if both clusters are empty, then merged node is empty */
    {
      clear_clust_stats(mgnode, x_len, y_len);
      mgnode->nb_points=0;
      mgnode->pred_err=0.0;
    }

}



static double abval_sq(double *vec,int x_len)    
{
  int i;
  double out=0.0;

  for (i=0;i<x_len;i++)
    out += (vec[i]*vec[i]);

  return(out);
}







/* Routine below can be used for debugging, commented out */

/*double comp_mergecst_slow(struct node *clust,struct data *train, struct cmds *params, struct node *mgnode, int cl1, int cl2)
{
  int nb_points,nb_clusters,x_len, y_len;
  double mse_cl1, mse_cl2, mse_mg;
  int i,j,cl,k;
  double y_hat_j, diff;

  nb_clusters = params->nb_clusters;
  nb_points=train->nb_points;
  x_len = train->x_len; 
  y_len = train->y_len; 
  
  mse_cl1 = 0.0;
  mse_cl2 = 0.0;
  mse_mg  = 0.0;

  for (i=0;i<nb_points;i++)
    {
      cl=determine_node(train->data[i], x_len, clust, params);
      
      if (cl==cl1)
	for (j=0;j<y_len;j++)
	  {
	    y_hat_j=0;


	    for (k=0;k<x_len;k++)
	      y_hat_j +=(train->data[i][k]  * clust[cl1].a[k][j]);  



	    y_hat_j += clust[cl1].b[j];
 

	    mse_cl1 += ((y_hat_j - train->data[i][j+x_len])*(y_hat_j - train->data[i][j+x_len]));
	    
	  }
      

      if (cl==cl2)
	for (j=0;j<y_len;j++)
	  {
	    y_hat_j=0;
	    for (k=0;k<x_len;k++)
	      y_hat_j +=(train->data[i][k]  * clust[cl2].a[k][j]);  

	    y_hat_j += clust[cl2].b[j];
    


	    mse_cl2 += ((y_hat_j - train->data[i][j+x_len])*(y_hat_j - train->data[i][j+x_len]));
	   

	  }


      if ((cl==cl1) || (cl==cl2))
	for (j=0;j<y_len;j++)
	  {
	    y_hat_j=0;
	    for (k=0;k<x_len;k++)
	      y_hat_j +=(train->data[i][k] * mgnode->a[k][j]);  

	    y_hat_j += mgnode->b[j];
	    
	    mse_mg += ((y_hat_j - train->data[i][j+x_len])*(y_hat_j - train->data[i][j+x_len]));
	    
	  }
    }
  
  diff = mse_mg-( mse_cl1+mse_cl2 );

  return(diff);
}
*/ 


